/*-
 * Copyright (C) @BABOLO  2002 May 01
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: babololib.c,v 1.1 2002/07/03 02:05:59 babolo Exp $ 
 */

#include <sys/types.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#ifdef DEBUG
# include <stdio.h>
#endif
#include "babololib.h"

int
readetheraddr( char **ethertext
             , u_int64_t *etheraddr
             )
/* reads *ethertext as MAC address and converts to u_int_64 */
{   u_int64_t j;
    u_int32_t i, k;
    char run;

    i = 0; k = 0; *etheraddr = 0; j = 0;
    for (run = 1; run; (*ethertext)++)
    {   switch (**ethertext)
        {case '0':case '1':case '2':case '3':case '4':case '5':case '6':case '7':case '8':case '9':
            if  (++i > 12)                                                                  return(13);
            k++;
            j <<=4;
            j |= **ethertext - '0';
            break;
         case 'a':case 'b':case 'c':case 'd':case 'e':case 'f':
            if  (++i > 12)                                                                  return(14);
            k++;
            j <<=4;
            j |= **ethertext - 'a' + 10;
            break;
         case 'A':case 'B':case 'C':case 'D':case 'E':case 'F':
            if  (++i > 12)                                                                  return(15);
            k++;
            j <<=4;
            j |= **ethertext - 'A' + 10;
            break;
         case ':':
            run++;
            /* break through */
         default:
            run--;
            i += k & 1;
            k += k & 1;
            *etheraddr <<= k * 4;
            *etheraddr |= j;
            k = 0;
            j = 0;
    }   }
    return((i == 12) ? 0 : i + 1);
/* return codes
 *  0 - OK
 *  1 - ether addr  0 long
 *  2 - ether addr  1 long
 *  3 - ether addr  2 long
 *  4 - ether addr  3 long
 *  5 - ether addr  4 long
 *  6 - ether addr  5 long
 *  7 - ether addr  6 long
 *  8 - ether addr  7 long
 *  9 - ether addr  8 long
 * 10 - ether addr  9 long
 * 11 - ether addr 10 long
 * 12 - ether addr 11 long
 * 13 - ether addr longer then 6 bytes
 * 14 - ether addr longer then 6 bytes
 * 15 - ether addr longer then 6 bytes
 */
}

int
readipaddr( char **iptext
          , u_int32_t *ipaddr
          )
/* reads in as IP address and converts to u_int_32 */
{   long long j;
    u_int32_t i;
    long k;
    char *inn = NULL;

    j = 0LL; *ipaddr = 0;
    for (i = 0; i < 4;)
    {   if  (i)
        {   if  (**iptext == '.') (*iptext)++;
            else if (strspn(*iptext, " \t\r\n") == 0)                                       return(33);
            else break;
        }else if (strspn(*iptext, "0123456789") == 0)                                       return(34);
        inn = *iptext;
        j = strtoll(*iptext, iptext, 0);
        if  (  j == LLONG_MAX || j == LLONG_MIN)                                            return(-1);
        if  (  (j & 0xffffffff00000000LL) != ((j < 0LL) ? 0xffffffff00000000LL : 0LL)
            || *iptext == inn
            )                                                                               return(35);
        j &= 0x00000000ffffffffLL;
        if  (j < 0x0100LL)
        {   *ipaddr <<= 8;
            *ipaddr |= (u_int32_t)j;
            i += 1;
        }else if (j < 0x010000LL)
        {   *ipaddr <<= 16;
            *ipaddr |= (u_int32_t)j;
            i += 2;
        }else if (j < 0x01000000LL)
        {   *ipaddr <<= 24;
            *ipaddr |= (u_int32_t)j;
            i += 3;
        }else
        {   *ipaddr = (u_int32_t)j;
            i += 4;
        }
#ifdef DEBUG
        fprintf( stderr, "%1d, %08X, %08X, %s~\n", i, *ipaddr, (u_int32_t)j, *iptext);
#endif
    }
    if  (i > 4)                                                                             return(36);
    i = 4 - i;
    switch (i)
    {case 1:
        *ipaddr <<= 8;
        if   (j < 0x0100LL)        *ipaddr &= 0xffff0000;
        else if (j < 0x010000LL)   *ipaddr &= 0xff000000;
        else if (j < 0x01000000LL) *ipaddr = 0;
        *ipaddr |= j;
        break;
     case 2:
        *ipaddr <<= 16;
        if   (j < 0x0100LL)      *ipaddr &= 0xff000000;
        else if (j < 0x010000LL) *ipaddr = 0;
        *ipaddr |= j;
        break;
    }
    k = 32L;
    if  (**iptext == '/')
    {   (*iptext)++;
        inn = *iptext;
        k = strtol(*iptext, iptext, 0);
        if  (  k == LONG_MAX || k == LONG_MIN)                                              return(-2);
        if  (k < 0L || k > 32L || *iptext == inn)                                           return(38);
    }else if (**iptext) return(37);
    return(k);
/* return codes
 * -2 - strtol (mask)
 * -1 - strtoll (*ipaddr)
 * 0 .. 32 result mask length
 * 33 - Illegal syntax
 * 34 - Illegal syntax
 * 35 - Out of range
 * 36 - Out of range
 * 37 -    
 * 38 - Mask out of range
 */
}

const struct ftree_node ftree_null = { 0, 0, 0, 0};

const struct ftree_node *
ftree_resolve( const struct ftree_node ftree[]
             , u_int32_t ipaddr
             )
{   int p;

    if  (!ftree) return(&ftree_null);

    for (p = 0; ftree[p].mask != 0;)
    {   if  ((ftree[p].mask & ipaddr) == ftree[p].value) p = ftree[p].eq;
          else p = ftree[p].neq;
    }

    return(&ftree[p]);
}
