/*-
 * Copyright (C) @BABOLO  2000 Jul 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: ra-demultiplex.c,v 1.7 2003/09/25 09:56:37 babolo Exp $ 
 */

#include <sys/syslimits.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <err.h>

#define HOST_BITS_PER_INT 32                     /* borrowed from argus (Carnegie Mellon University) */
#define argtimeval timeval

#if defined(__sgi) || defined(HAVE_SOLARIS) || defined(ultrix) || defined(__osf__) || defined(linux) || defined(AIX)

#define timelocal mktime

#if defined(__sgi)
#include <bstring.h>
#include <ulimit.h>

#if _MIPS_SZLONG == 64
#undef argtimeval
#define argtimeval irix5_timeval
#endif

#undef TCPSTATES
struct ether_addr
{   u_char  octet[6];
};
#endif
#endif

#if defined(AIX)
#define _SUN
#include <sys/select.h>
#include <net/nh.h>
#endif

#define arg_uint8   u_char
#define arg_int8    char
#define arg_uint16  u_short
#define arg_int16   short

#if HOST_BITS_PER_INT == 32
#define arg_uint32  u_int
#define arg_int32   int
#else
#define arg_uint32  u_long
#define arg_int32   long
#endif                                       /* END borrowed from argus (Carnegie Mellon University) */

#include <cons_def.h>
#include <cons_out.h>
#include "bufread.h"

#define IP_ADDR_LEN 4
#define USE_IP 0x01000000000000
#define omode ((mode_t)0777)

struct ef_entry
{   u_int64_t ea;
    u_int32_t ia;
    int file;
    unsigned counter, ocount;
    int next;
};

int flag_v;
int openmax;
struct ef_entry *ef_array;
int lhead, ltail;
u_char *swdir, *swfile, *swname, *suffix;
struct WriteStruct head, next;
u_int64_t *maclist;
unsigned maclen;
unsigned count;

u_int64_t eatoll(struct ether_addr *ea)
{   int i;
    u_int64_t ll;
    /* Convert ether_addr to u_int64_t */
    for (i = 0, ll = 0; i < ETHER_ADDR_LEN; i++)
    {   ll <<= 8;
        ll |= ea->octet[i];
    }
    /* Does IP need to be part of filename */
    if  (suffix)
    {   unsigned probe, mask = 1;
        /* Binary tree seek */
        for (mask = 1; mask < maclen; mask <<= 1);
        if  (mask >= maclen) mask >>= 1;
        for (probe = mask; mask; probe |= mask)
        {   if  (ll < maclist[probe]) probe &= ~mask;
            mask >>= 1;
        }
        if  (maclen == 0 || ll != maclist[probe]) ll |= USE_IP;
    }
    return(ll);
}

void prentry_debug(char *s, u_int64_t ll, u_int32_t ii)
{   fprintf( stderr, "%s %02x %02x:%02x:%02x:%02x:%02x:%02x %u.%u.%u.%u\n"
           , s
           , (unsigned)(ll >> 48) & 0377
           , (unsigned)(ll >> 40) & 0377
           , (unsigned)(ll >> 32) & 0377
           , (unsigned)(ll >> 24) & 0377
           , (unsigned)(ll >> 16) & 0377
           , (unsigned)(ll >> 8) & 0377
           , (unsigned)(ll) & 0377
           , (ii >> 24) & 0377, (ii >> 16) & 0377
           , (ii >> 8) & 0377, ii & 0377
           );
}

int find_file(u_int64_t ll, u_int32_t ii)
{   int i;
    if  (flag_v > 1) prentry_debug("f", ll, ii);
    for (i = 0; i < openmax; i++)
    {   if  (flag_v > 2) prentry_debug("l", ef_array[i].ea, ef_array[i].ia);
        if  (   ef_array[i].ea == ll
            &&  (   !(ll & USE_IP)
                ||  ef_array[i].ia == ii
            )   )
            return(i);
    }
    return(-1);
}

/* Sorts output files for LRU to substitute */
int sort(int beg, int len)
{   int res;
    switch (len)
    {case 0:
        res = -1;
        break;
     case 1:
        res = beg;
        ef_array[beg].next = -1;
        break;
     /* May be case 2 & 3 for eficiency here */
     default:
        {   int list0, list1, len0, *p;
            len0 = len / 2;
            list0 = sort(beg, len0);
            list1 = sort(beg + len0, len - len0);
            res = -1;
            p = &res;
            while (list0 >= 0 || list1 >= 0)
            {   if  (list0 == -1)
                {   *p = list1;
                    break;
                }else if (list1 == -1)
                {   *p = list0;
                    break;
                }else if (ef_array[list0].counter < ef_array[list1].counter)
                {
                    *p = list0;
                    p = &ef_array[*p].next;
                    list0 = *p;
                }else
                {
                    *p = list1;
                    p = &ef_array[*p].next;
                    list1 = *p;
    }   }   }   }
    return(res);
}

int free_file()
{   int i;
                /* Get clear entry if possible. Possible in the very beginning until enties polluted */
    if  (lhead >= 0 && ef_array[lhead].file == 0)
    {   i = lhead;
        lhead = ef_array[lhead].next;
        if  (flag_v > 1) fprintf(stderr, "Got clear entry %d\n", i);
        return(i);
    }
                                                      /* Get oldest entry from the oldets entry list */
    while (lhead >= 0)
    {   i = lhead;
        lhead = ef_array[lhead].next;
        if  (flag_v > 1) fprintf(stderr, "Test %d\n", i);
        if  (ef_array[i].counter == ef_array[i].ocount)
        {   if  (flag_v > 1) fprintf(stderr, "Got old entry %d\n", i);
            close(ef_array[i].file);
            return(i);
        }
        ef_array[i].next = -1;
    }
                                                                 /* Oldets entry list empty. Fill it */
    if  (flag_v > 1) fprintf(stderr, "###### Populate oldest entry list\n");
    lhead = sort(0, openmax);
    count = 0;
    for (i = lhead; i >= 0; i = ef_array[i].next) ef_array[i].ocount = ef_array[i].counter = ++count;
    if (flag_v > 3) for (i = 0; i < openmax; i++) fprintf(stderr, "#%d next=%d count=%u\n", i, ef_array[i].next, ef_array[i].counter);
    i = lhead;
    lhead = ef_array[i].next;
    close(ef_array[i].file);
    if  (flag_v > 1) fprintf(stderr, "Got old entry %d\n", i);
    return(i);
}

void re_count()                                                       /* How to test it? Test if can */
{   int i;
    if  (flag_v) fprintf(stderr, "Recount now\n");
    lhead = sort(0, openmax);
    count = 0;
    for (i = lhead; i >= 0; i = ef_array[i].next) ef_array[i].ocount = ef_array[i].counter = ++count;
}

void do_rec(struct ether_addr *ea, struct in_addr *ia)
{   u_int64_t ll = eatoll(ea);
    u_int32_t ii = ntohl(ia->s_addr);
    int indx;
    int filn;
    if  (count + 1 == 0) re_count();
    if  ((indx = find_file(ll, ii)) < 0)
    {   indx = free_file();
        if  (ll & USE_IP)
            sprintf( swname, "%02x:%02x:%02x:%02x:%02x:%02x%s%u.%u.%u.%u"
                   , ea->octet[0], ea->octet[1]
                   , ea->octet[2], ea->octet[3]
                   , ea->octet[4], ea->octet[5]
                   , suffix
                   , ((u_char*)ia)[0], ((u_char*)ia)[1]
                   , ((u_char*)ia)[2], ((u_char*)ia)[3]
                   );
        else
            sprintf( swname, "%02x:%02x:%02x:%02x:%02x:%02x"
                   , ea->octet[0], ea->octet[1]
                   , ea->octet[2], ea->octet[3]
                   , ea->octet[4], ea->octet[5]
                   );
        if  ((filn = open(swfile, O_WRONLY | O_APPEND)) < 0)
        {   if  (errno != ENOENT)                             err(EX_IOERR, "can't create %s", swfile);
            if  ((filn = open(swfile, O_WRONLY | O_APPEND | O_CREAT, 0666)) < 0)
                                                              err(EX_IOERR, "can't create %s", swfile);
            if  (flag_v > 2) fprintf(stderr, "New ");
            if  (sizeof(head) != write(filn, &head, sizeof(head)))
                                                              err( EX_IOERR, "can't write %s", swfile);
        };
        ef_array[indx].file = filn;
        if  (flag_v > 1) fprintf(stderr, "Entry %d file #%d %s\n", indx, filn, swname);
    };
    ef_array[indx].ea = ll;
    ef_array[indx].ia = ii;
    ef_array[indx].next = -1;
    ef_array[indx].counter = ++count;
    if  (flag_v > 2)
        fprintf(stderr, "Write to %d file #%d(%u)\n", indx, ef_array[indx].file, count);
    if  (sizeof(next) != write(ef_array[indx].file, &next, sizeof(next)))
                                                err( EX_IOERR, "write to %02x:%02x:%02x:%02x:%02x:%02x"
                                                   , ea->octet[0], ea->octet[1]
                                                   , ea->octet[2], ea->octet[3]
                                                   , ea->octet[4], ea->octet[5]
                                                   );
}

void build(char *gpath)                                          /* borrowed from FreeBSD's mkdir(1) */
{   struct stat sb;
    int last;
    u_char *p, *path;

    if  (!(path = strdup(gpath)))                                            err(EX_CONFIG, "Low mem");
    p = path;
    if  (p[0] == '/') ++p;
    for (last = 0; !last ; ++p)
    {   if  (p[0] == '\0') last = 1;
        else if (p[0] != '/') continue;
        *p = '\0';
        if  (p[1] == '\0') last = 1;
        if  (stat(path, &sb))
        {   if (errno != ENOENT || mkdir(path, omode) < 0)               err(EX_CANTCREAT, "%s", path);
            else if (flag_v) fprintf(stderr, "%s\n", path);
        }else if ((sb.st_mode & S_IFMT) != S_IFDIR)
        {   if  (last) errno = EEXIST;
            else errno = ENOTDIR;
                                                                         err(EX_CANTCREAT, "%s", path);
        }
        if  (!last) *p = '/';
    }
    free(path);
}

void usage(int status)
{   fprintf( stderr
           , "ra-demultiplex @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "ra-demultiplex [-r infile] [-d outprefix]\n"
           );
    exit(status);
}

void readMAClist(u_char *macfile)
{   FILE *f;
    int fd = open(macfile, O_RDONLY | O_SHLOCK);
    struct stat sb;
    u_int64_t m, mm;
    int c, n, nl, nm, mll, p;

    if  (fd < 0)                                            err(EX_NOINPUT, "Can't open: %s", macfile);
    if  (fstat(fd, &sb) < 0)                               err(EX_NOINPUT, "Can't fstat: %s", macfile);
    f = fdopen(fd, "r");
    if  (!f)                                               err(EX_NOINPUT, "Can't fopen: %s", macfile);
    mll = (sb.st_size / 12 + 513) & ~0x1FF;              /* 12 is minimum MAC length (0:0:0:0:0:0\n) */
    if  ((maclist = malloc(mll * sizeof(u_int64_t))) == NULL) err(EX_CONFIG, "Low mem in readMAClist");
    maclen = 0;
    n = 0;                                              /* hexadecimal digit number in current tuple */
    m = 0;                                                                 /* value of current tuple */
    c = 0;
    nl = 1;                                                                /* input file line number */
    nm = 0;                                                           /* digit number in MAC address */
    mm = 0;                                                                           /* MAC address */
    while (1)
    {   switch (c = fgetc(f))
        {case '0': case '1': case '2': case '3': case '4':
         case '5': case '6': case '7': case '8': case '9':
    ;       m <<= 4;
    ;       m += (c - '0');
    ;       n++;
    ;       break;
         case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
    ;       m <<= 4;
    ;       m += (c + 10 - 'a');
    ;       n++;
    ;       break;
         case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
    ;       m <<= 4;
    ;       m += (c + 10 - 'A');
    ;       n++;
    ;       break;
         case ':':
    ;       mm <<= (((n + 1) & ~1) * 4);
    ;       nm += (n + 1) & ~1;
    ;       mm |= m;
    ;       m = 0;
    ;       n = 0;
    ;       break;
         case '#':
    ;       do
            {   c = fgetc(f);
    ;       } while (c > '0' && c != '\n');
         case '\n':
    ;       nl++;
         default:
    ;       if  (n)
            {   mm <<= (((n + 1) & ~1) * 4);
    ;           nm += (n + 1) & ~1;
    ;           mm |= m;
    ;           m = 0;
    ;           n = 0;
    ;           if  (c > ' ' || nm != 12)
                {   fprintf(stderr, "Macaddr error in string %d of file %s\n", nl, macfile);
    ;               exit(EX_DATAERR);
                }
    ;           /* Insert mm into sorted list *ml */
    ;           p = ++maclen;
    ;           if  (mll < p)
                {   mll += 512;
    ;               if  ((maclist = reallocf(maclist, mll * sizeof(u_int64_t))) == NULL)
                                                              err(EX_CONFIG, "Low mem in readMAClist");
    ;           }
    ;           /* bubble sort - onse per MAC list */
    ;           while (--p > 0 && mm < maclist[p]) maclist[p + 1] = maclist[p];
    ;           maclist[p] = mm;
    ;           mm = 0;
    ;           nm = 0;
    ;       }
    ;       if  (c < 0) /* EOF */
            {   if  (flag_v)
                    for (p = 0; p < maclen; p++)
                        fprintf( stderr, "%02x:%02x:%02x:%02x:%02x:%02x\n"
                               , (unsigned)(maclist[p] >> 40) & 0377
                               , (unsigned)(maclist[p] >> 32) & 0377
                               , (unsigned)(maclist[p] >> 24) & 0377
                               , (unsigned)(maclist[p] >> 16) & 0377
                               , (unsigned)(maclist[p] >> 8) & 0377
                               , (unsigned)(maclist[p]) & 0377
                               );
    ;           return;
}   }   }   }

int main(int argc, char **argv)
{   int i, c;
    int fin;
    u_int64_t n;
    u_char *p;
    size_t swlen;

    flag_v = 0;
    fin = fileno(stdin);
    swdir = ".";
    suffix = NULL;
    maclen = 0;
    while ((c = getopt(argc, argv, "vr:d:m:s:h")) > 0)
    {   switch (c)
        {case 'v': flag_v++;                                 break;
         case 'r':
            if  (!(fin = open(optarg, O_RDONLY)))            err(EX_NOINPUT, "Can't open: %s", optarg);
                                                             break;
         case 'd': swdir = optarg;                           break;
         case 'm':
            readMAClist(optarg);
            if  (suffix == NULL) suffix = "+";               break;
         case 's':
            suffix = optarg;
                                                             break;
         case 'h': usage(EX_OK);
         default: usage(EX_USAGE);
    }   }

    argc -= optind;
    argv += optind;

    if  ((p = rindex(swdir, '/')))
    {   *p = '\0';
        if  (flag_v > 1) fprintf(stderr, "Build %s~\n", swdir);
        build(swdir);
        *p = '/';
    }
    count = 0;

    /* openmax = OPEN_MAX - 4; */
    openmax = getdtablesize() / 2 - 5;
    if  (openmax < OPEN_MAX) openmax = OPEN_MAX;

    ef_array = (struct ef_entry *)malloc(sizeof(struct ef_entry) * openmax);
    if  (!ef_array)                                                          err(EX_CONFIG, "Low mem");
    bzero(ef_array, sizeof(struct ef_entry) * openmax);
    if  (flag_v > 1) fprintf(stderr, "%d output files\n", openmax);
    for (i = 0; i < openmax; i++) ef_array[i].next = i + 1;
    lhead = 0;
    ltail = openmax - 1;
    ef_array[ltail].next = -1;

    swlen = 4 + strlen(swdir) + ETHER_ADDR_LEN * 3;
    if  (suffix) swlen += strlen(suffix) + IP_ADDR_LEN * 4;
    if  (flag_v > 1) fprintf(stderr, "Total output filename length=%u\n", swlen);
    swfile = (char *)malloc(swlen);
    if  (!swfile)                                                            err(EX_CONFIG, "Low mem");
    bzero(swfile, swlen);
    strcpy(swfile, swdir);
    swname = index(swfile, '\0');

    init_next();
    for (n = 0; get_next(fin, &next) == sizeof(next); n++)
    {   if  (n == 0)
        {   if  (next.status & htonl(ARGUSCONTROL))
                bcopy(&next, &head, sizeof(next));
            else
            {   fprintf(stderr, "data error\n");                                      exit(EX_DATAERR);
            }
        }else if (next.status & htonl(IPPROTO))
        {   do_rec(&next.ws_ip_phys.ethersrc, &next.ws_ip_src);
            do_rec(&next.ws_ip_phys.etherdst, &next.ws_ip_dst);
    }   }
    for (i = 0; i < openmax; i++) if (ef_array[i].file) (void)close(ef_array[i].file);
    exit(EX_OK);
}
